/*
 * iofe.c: Io frontend.
 * Copyright (c) 2004 Kentaro A. Kurahone.  All Rights Reserved.
 * 
 * Portions derived of IoVM code by Steve Dekorte et al.
 */

#include <stdlib.h>
#include <stdio.h>
#include <histedit.h>

#include "IoState.h"
#include "IoFile.h"
#include "IoList.h"
#include "IoNumber.h"
#include "IoDirectory.h"

#include "IoEditLine.h"
#include "IofeCompletion.h"

#define IOFE_VERSION 20070601
#define NEWLINE_LEN     2

/* Global variables. */
History *histCtxt = NULL;
EditLine *editCtxt = NULL;
int histSize = 0;
int isMulti = 0;

/* File global variables. */
static const char fePrompt[] = "Io> ";
static const char feMultiPrompt[] = "> ";

/* Static functions. */
static char* promptCallback(EditLine *edit);
static void printCallback(void *state, const size_t count, const char *s);
static void exceptionCallback(void *state, IoObject *e);
static void exitCallback(void *state, int dummy); /* @@@ is it dummy? */
static void doInteractive(IoState *state);
static void doFile(IoState *state, int argc, const char *argv[]);
static char* doGetInput(void);

static char* promptCallback(EditLine *edit)
{
  return (isMulti) ? (char *) feMultiPrompt : (char *) fePrompt;
}

static void printCallback(void *state, const size_t count, const char *s)
{
  fwrite(s, 1, count, stdout);
}

static void exceptionCallback(void *state, IoObject *e)
{
  IoCoroutine_rawPrintBackTrace(e);
}

static void exitCallback(void *state, int dummy)
{
  /* IoState_free(state); */
  history_end(histCtxt);
  el_end(editCtxt);
  exit(0);
}

static char* doGetInput(void)
{
  char *rval = NULL, *buf = NULL, *tmp = NULL;
  int done = 0, count;

  while (!done) {
    buf = (char *) el_gets(editCtxt, &count);
    if (buf) {
      if (buf[count - NEWLINE_LEN] == 0x5C) {
        isMulti = 1;
        /* Multi line input. */
        if (rval == NULL) {
          rval = (char *) malloc(strlen(buf) + 1);
          strncpy(rval, buf, strlen(buf));
          rval[count - NEWLINE_LEN] = '\0';
        } else {
          tmp = realloc(rval, strlen(rval) + count);
          if (tmp) {
            rval = tmp;
            strncat(rval, buf, count - NEWLINE_LEN);
          } else {
            /* Out of memory. */
            fprintf(stderr, "Out of memory. Aborting.\n");
            exit(-1);
          }
        }
      } else if(isMulti) {
        /* End of multi line input. */
        tmp = realloc(rval, strlen(rval) + count + 1);
        if (tmp) {
          rval = tmp;
          strncat(rval, buf, count);
          done = 1;
        } else {
          /* Out of memory. */
          fprintf(stderr, "Out of memory. Aborting.\n");
          exit(-1);
        }
      } else {
        /* Only one line. */
        rval = (char *) malloc(strlen(buf) + 1);
        strncpy(rval, buf, strlen(buf) + 1);
        done = 1;
      }
    } else {
      done = 1;
    }
  }

  isMulti = 0;

  return rval;
}

static void doInteractive(IoState *state)
{
  IoObject *lobby = state->lobby;
  IoObject *result;
  IoObject *iofeObj;
  UArray *path;
  char *buf;
  HistEvent ev;

  /* Set the launchpath. */
  path = IoDirectory_CurrentWorkingDirectoryAsUArray();
  IoObject_setSlot_to_(lobby,
    IoState_symbolWithCString_(state, "launchPath"),
    IoSeq_newWithUArray_copy_(state, path, 0));

  /* New IoFE object. */
  iofeObj = IoState_doCString_(state, "IoFE := Object clone");
  newEditLineObject(state, lobby);
  initCompletion(state, iofeObj);

  /* Set completion hook. */
  el_set(editCtxt, EL_ADDFN, "iofe-complete", "Completion", invokeCompletion);
  el_set(editCtxt, EL_BIND, "\t", "iofe-complete", NULL);

  /* Interactive mode. */
  while(1) {
    buf = doGetInput();
    if (buf) {
      result = IoState_doCString_(state, buf);
      if (result) {
        history(histCtxt, &ev, H_ENTER, buf);
        fprintf(stdout, "==> ");
        IoObject_print(result);
        fprintf(stdout, "\n");
      }
      free(buf);
    } else {
      /* Invoke exit code on Ctrl-D. */
      fprintf(stdout, "\n");
      exitCallback(state, 0);
    }
  }
}

static void doFile(IoState *state, int argc, const char *argv[])
{
  IoObject *lobby = state->lobby;
  IoList *argsList = IoList_new(state);
  char *filePath = (char *) argv[1];
  UArray *path;
  int i;

  /* Add argc/argv to the VM. */
  if (argc >= 2) {
    for (i = 2; i < argc; i ++) {
      IoList_rawAppend_(argsList,
        IoSeq_newWithCString_(state, (char *) argv[i]));
    }
    IoObject_setSlot_to_(lobby,
      IoState_symbolWithCString_(state, "args"),
      argsList);
  }

  /* Extract the launchPath from the file. */
  if (*filePath == '/' || *filePath == '\\') {
    path = UArray_newWithCString_(filePath);
  }
  else
  {
    /* @@@ use IoDirectory */
    path = IoDirectory_CurrentWorkingDirectoryAsUArray();
    UArray_appendPath_(path, UArray_newWithCString_(filePath));
  }

  UArray_removeLastPathComponent(path);

  IoObject_setSlot_to_(lobby,
                       IoState_symbolWithCString_(state, "launchPath"),
                       IoSeq_newWithUArray_copy_(state, path, 0));

  /* Run the VM. */
  IoState_doFile_(state, filePath);
//  while(IoState_activeCoroCount(state) > 1) {
//    IoState_yield(state);
//  }
}

int main(int argc, const char *argv[])
{
  IoState *state;
  IoNumber *version;
  IoObject *lobby;
  IoObject *sys;
  HistEvent ev;

  /* Initialise libedit. */
  editCtxt = el_init(argv[0], stdin, stdout, stderr);
  histCtxt = history_init();
  history(histCtxt, &ev, H_SETSIZE, 10);
  el_set(editCtxt, EL_PROMPT, promptCallback);
  el_set(editCtxt, EL_HIST, history, histCtxt);
  histSize = 10; /* HACK: Need this for EditLine obj. */
  /*
   * HACK: EditLine takes care of this. 
   * el_set(editCtxt, EL_EDITOR, "emacs");
   */

  /* Initialise IoVM. */
  state = IoState_new();
  IoState_callbackContext_(state, NULL);
  IoState_printCallback_(state, printCallback);
  IoState_exceptionCallback_(state, exceptionCallback);
  IoState_exitCallback_(state, exitCallback);
  IoState_argc_argv_(state, argc, argv);

  /* Extract lobby. */
  lobby = state->lobby;
  sys = IoObject_getSlot_(lobby, IoState_symbolWithCString_(state, "System"));
  version = IoObject_getSlot_(sys,
    IoState_symbolWithCString_(state, "version"));

  /* Kick off the VM. */
  if (argc >= 2) {
    /* If there is a file specified, run it. */
    doFile(state, argc, argv);
  } else { 
    /* Print banner. */
    IoState_print_(state, "IoFE version %d (Io version %d)\n",
                   IOFE_VERSION,
                   IoNumber_asInt(version));

    /* Run in interactive mode. */
    doInteractive(state);
  }

  return 0;
}
